use object::{
	ship::{OutfitSet, ShipPersonality},
	GameData, GameShipHandle,
};
use std::time::Instant;
use winit::event::{ElementState, MouseButton, MouseScrollDelta, TouchPhase, VirtualKeyCode};

use crate::camera::Camera;
use crate::inputstatus::InputStatus;

use galactica_constants;
use galactica_content as content;
use galactica_gameobject as object;
use galactica_render::RenderState;
use galactica_world::{objects::ShipControls, ParticleBuilder, StepResources, World};

pub struct Game {
	input: InputStatus,
	last_update: Instant,
	player: GameShipHandle,
	paused: bool,
	time_scale: f32,
	start_instant: Instant,
	camera: Camera,

	// TODO: include system in world
	//system: object::System,
	gamedata: GameData,

	content: content::Content,
	world: World,
	new_particles: Vec<ParticleBuilder>,
}

impl Game {
	pub fn new(ct: content::Content) -> Self {
		let mut gamedata = GameData::new(&ct);

		let ss = ct.get_ship(content::ShipHandle { index: 0 });

		let mut o1 = OutfitSet::new(ss.space, &[]);
		o1.add(&ct.get_outfit(content::OutfitHandle { index: 0 }));
		o1.add(&ct.get_outfit(content::OutfitHandle { index: 1 }));
		//o1.add_gun(&ct, content::GunHandle { index: 0 });
		//o1.add_gun(&ct, content::GunHandle { index: 0 });
		//o1.add_gun(&ct, content::GunHandle { index: 0 });

		print!("{:?}", o1);
		let player = gamedata.create_ship(
			&ct,
			content::ShipHandle { index: 0 },
			content::FactionHandle { index: 0 },
			ShipPersonality::Player,
			o1.clone(),
			&content::SystemHandle { index: 0 },
		);

		gamedata.create_ship(
			&ct,
			content::ShipHandle { index: 0 },
			content::FactionHandle { index: 1 },
			ShipPersonality::Dummy,
			OutfitSet::new(ss.space, &[]),
			&content::SystemHandle { index: 0 },
		);

		let mut o1 = OutfitSet::new(ss.space, &[]);
		o1.add(&ct.get_outfit(content::OutfitHandle { index: 0 }));
		//o1.add_gun(&ct, content::GunHandle { index: 0 });

		gamedata.create_ship(
			&ct,
			content::ShipHandle { index: 0 },
			content::FactionHandle { index: 1 },
			ShipPersonality::Point,
			o1,
			&content::SystemHandle { index: 0 },
		);

		let physics = World::new(&ct, &gamedata, content::SystemHandle { index: 0 });

		Game {
			last_update: Instant::now(),
			input: InputStatus::new(),
			player,
			start_instant: Instant::now(),

			camera: Camera {
				pos: (0.0, 0.0).into(),
				zoom: 500.0,
				aspect: 1.0,
			},
			//system: object::System::new(&ct, SystemHandle { index: 0 }),
			paused: false,
			time_scale: 1.0,
			world: physics,
			gamedata,
			content: ct,
			new_particles: Vec::new(),
		}
	}

	pub fn set_camera_aspect(&mut self, v: f32) {
		self.camera.aspect = v
	}

	pub fn process_key(&mut self, state: &ElementState, key: &VirtualKeyCode) {
		self.input.process_key(state, key)
	}

	pub fn process_click(&mut self, state: &ElementState, key: &MouseButton) {
		self.input.process_click(state, key)
	}

	pub fn process_scroll(&mut self, delta: &MouseScrollDelta, phase: &TouchPhase) {
		self.input.process_scroll(delta, phase)
	}

	pub fn set_paused(&mut self, pause: bool) {
		if pause {
			self.paused = true;
			self.input.release_all()
		} else {
			self.paused = false;
		}
	}

	pub fn update(&mut self) {
		let t: f32 = self.last_update.elapsed().as_secs_f32() * self.time_scale;

		let world_output = self.world.step(StepResources {
			player: self.player,
			player_controls: ShipControls {
				left: self.input.key_left,
				right: self.input.key_right,
				thrust: self.input.key_thrust,
				guns: self.input.key_guns,
			},
			ct: &self.content,
			dt: &mut self.gamedata,
			particles: &mut self.new_particles,
			t,
		});

		if self.input.v_scroll != 0.0 {
			self.camera.zoom = (self.camera.zoom + self.input.v_scroll)
				.clamp(galactica_constants::ZOOM_MIN, galactica_constants::ZOOM_MAX);
			self.input.v_scroll = 0.0;
		}

		self.camera.pos = world_output.player_position;
		self.last_update = Instant::now();
	}

	pub fn get_frame_state(&mut self) -> RenderState {
		RenderState {
			camera_pos: self.camera.pos,
			camera_zoom: self.camera.zoom,
			current_time: self.start_instant.elapsed().as_secs_f32(),
			content: &self.content,
			world: &self.world,
			particles: &mut self.new_particles,
			player_data: self.player,
			data: &self.gamedata,
		}
	}
}
